

from itertools import groupby, count
from operator import itemgetter

from datetime import timedelta, datetime
from epsilon.extime import Time
from epsilon import structlike

linkStyle = "text-decoration: none; color: white;"

from twisted.web.template import Element, XMLFile, renderer, tags, flatten

from twisted.python.modules import getModule

here = getModule(__name__).filePath

class HiscoresPage(Element):
    loader = XMLFile(here.sibling('hiscore.html').open())

    def __init__(self, dateWithinMonth, scores):
        self.dateWithinMonth = dateWithinMonth
        self.allScores = scores


    def linkTag(self, time, arrowDirection, hidden=False):
        style = linkStyle
        if hidden:
            style += ' visibility: hidden;'
        return tags.a(
            style=style,
            href="?time=%s" % (time.asDatetime().strftime('%Y-%m-%d'),)
        )(
            tags.img(border="0", src="%s-arrow.png" % (arrowDirection,))
        )


    @renderer
    def next(self, request, tag):
        start, end = monthRangeAround(self.dateWithinMonth + timedelta(days=45))
        hidden = False
        if start > Time():
            hidden = True
        return tag(self.linkTag(start, "right", hidden))


    @renderer
    def previous(self, request, tag):
        start, end = monthRangeAround(self.dateWithinMonth - timedelta(hours=1))
        return tag(self.linkTag(start, "left"))


    @renderer
    def scores(self, request, oneRowTag):
        for pos, (score, author) in enumerate(self.allScores):
            pos += 1
            if pos == 1:
                color = 'yellow'
            elif pos == 2:
                color = 'white'
            elif pos == 3:
                color = 'orange'
            else:
                if pos > 10:
                    color = '#555555'
                else:
                    color = '#888888'
            rank = str(pos)
            if (pos % 100) > 10 and (pos % 100) < 20:
                suffix = 'th'
            elif rank[-1] == '1':
                suffix = 'st'
            elif rank[-1] == '2':
                suffix = 'nd'
            elif rank[-1] == '3':
                suffix = 'rd'
            else:
                suffix = 'th'
            yield oneRowTag.clone().fillSlots(rank=str(pos) + suffix,
                                              color=color,
                                              bang="!" * max(0, 4-pos),
                                              total=str(score),
                                              author=author)


    @renderer
    def footer(self, request, tag):
        for factor in sorted(factors.values(), key=lambda f: f.order):
            tag(str(factor.points), " points for ", factor.description, tags.br())
        return tag


def render(output, rangeStart, scores):
    """
    Write some score data to a callable of some kind.

    @param rangeStart: A L{Time} instance giving the beginning of the time range
        containing the given scores.

    @param scores: the scores data structure as returned by getscores.

    @param output: a callable that takes a C{str}; call it repeatedly to write
        the output.
    """
    page = HiscoresPage(rangeStart, scores)
    # Not requesting any asynchronous data, so we can just ignore the result of
    # this Deferred.
    flatten(None, page, output)


factors = {}

class Factor(structlike.record('order points description')):
    """
    A score thing.
    """



_nextorder = count().next

def deffactor(score, description):
    """
    Define a new factor with an associated score and textual english
    description.

    Suck it, ubuntu.  Use software in ENGLISH.
    """
    key = (object(), repr(description))
    factors[key] = Factor(_nextorder(), score, description)
    return key



SUBMIT = deffactor(200, "submitting a ticket for review!")
FIXED = deffactor(500, "closing a ticket!!")
REVIEWED = deffactor(1000, "a review!!!")
BAD_REVIEW = deffactor(
    -600, ["a bad review", tags.br(),"(reviewing a ticket that gets reopened)"])
REOPENED = deffactor(10, "reopening a ticket")
FIX_COMMIT = deffactor(15, "committing a fix from a contributor")



def getscores(actions):
    scores = {}
    for (action, ticket, author, time) in actions:
        scores[author] = scores.get(author, 0) + factors[action].points
    return sorted([(score, author) for (author, score) in scores.items()],
                   reverse=True)


def parsetime(s):
    y, m, d = map(int, s.split("-"))
    return Time.fromDatetime(datetime(year=y, month=m, day=d))



HOW_LONG_IS_A_MONTH_I_DONT_KNOW = 27

def monthRangeAround(t):
    beginningdt = t.asDatetime().replace(day=1, hour=0, minute=0, second=0,
                                         microsecond=0)
    td = timedelta(days=HOW_LONG_IS_A_MONTH_I_DONT_KNOW)
    while (beginningdt + td).month == beginningdt.month:
        td += timedelta(days=1)
    beginning = Time.fromDatetime(beginningdt)
    end = Time.fromDatetime(beginningdt + td)
    return (beginning, end)


def getChanges(start, end):
    try:
        import pgdb
    except:
        import psycopg2
        con = psycopg2.connect("host=/tmp dbname=trac user=trac-migration")
    else:
        con = pgdb.connect(":trac")


    cur = con.cursor()
    cur.execute("""
                select ticket, time, author, field, oldvalue, newvalue
                from
                    ticket_change
                where
                    time > %s and time < %s
                order by
                    time desc
                """, [start.asPOSIXTimestamp(),
                      end.asPOSIXTimestamp()])

    return cur.fetchall()


def splitKeywords(keywords):
    return keywords.split()


def getActions(changes):
    """
    Compute the actions represented by the given sequence of ticket change rows.

    @param changes: A sequence of changes from a trac ticket_change table,
    ordered by time (so as to be able to detect which change rows are part of
    the same logical change).
    """
    lastReview = {} # map ticket(int): username of last reviewer (unicode-ish)

    for time, localChanges in groupby(changes, itemgetter(1)):
        actions = []
        comment = None
        for ticket, time, author, field, oldvalue, newvalue in localChanges:
            author = author.lower()
            if field == 'keywords':
                old = splitKeywords(oldvalue)
                new = splitKeywords(newvalue)
                if 'review' in old and 'review' not in new:
                    lastReview[ticket] = author
                    actions.append((REVIEWED, author))
                elif 'review' in new and 'review' not in old:
                    actions.append((SUBMIT, author))
            elif field == 'resolution':
                if newvalue == 'fixed':
                    # discover the previous submitter / reviewer
                    actions.append((FIXED, None))
                elif oldvalue == 'fixed':
                    actions.append((REOPENED, author))
                    if ticket in lastReview:
                        actions.append((BAD_REVIEW, lastReview[ticket]))
            elif field == 'comment':
                comment = newvalue
        for (action, possibleAuthor) in actions:
            if action is FIXED:
                commentAuthors = parseAuthorFromComment(comment)
                if commentAuthors is None:
                    yield (action, ticket, author, time)
                else:
                    for commentAuthor in commentAuthors:
                        yield (action, ticket, commentAuthor, time)
                    if author not in commentAuthors:
                        yield (FIX_COMMIT, ticket, author, time)
            else:
                yield (action, ticket, possibleAuthor, time)



def parseAuthorFromComment(comment):
    """
    Parse an author from the mandated commit-message format for Twisted, e.g. a
    line that starts with 'Author: ' somewhere in the comment.

    Things fall apart, the center cannot hold, and sometimes, people fill out
    this field wrong, so this will return None if it can't find such a thing.
    """
    if comment is None:
        return None
    for line in comment.splitlines():
        for variant in ['author:', 'authors:']:
            if line.lower().startswith(variant):
                return [s.strip(" \t\n\r.") for s in
                        line.lower().split(':')[1].split(',')]
    return None



def main(start, end, output):
    render(output, start, getscores(getActions(getChanges(start, end))))

if __name__ == '__main__':
    import pprint
    start, end = monthRangeAround(Time())
    actions = list(getActions(getChanges(start, end)))
    pprint.pprint(actions)
    pprint.pprint(getscores(actions))

