
"""
Unit tests for L{hiscore}.
"""

from twisted.trial.unittest import TestCase

from hiscore import (getActions, parseAuthorFromComment, REVIEWED, FIXED,
                     getscores, SUBMIT, REOPENED, BAD_REVIEW, FIX_COMMIT,
                     factors)

class GetActionsTests(TestCase):
    """
    Tests for L{hiscore.getActions}, the function which produces score data
    from a raw stream of ticket changes.
    """
    def test_submitted(self):
        """
        If a ticket change adds C{"review"} to the keywords field, L{getActions}
        emits a C{"submitted"} action.
        """
        ticket = 1234
        author = u'alice'
        time = 4321
        row = (ticket, time, author, u'keywords', u'easy', u'easy review')
        actions = list(getActions([row]))
        self.assertEquals(actions, [(SUBMIT, ticket, author, time)])


    def test_reviewed(self):
        """
        If a ticket change removes C{"review"} from the keywords field,
        L{getActions} emits a C{"reviewed"} action.
        """
        ticket = 1234
        author = u'alice'
        time = 4321
        row = (ticket, time, author, u'keywords', u'easy review', u'easy')
        actions = list(getActions([row]))
        self.assertEquals(actions, [(REVIEWED, ticket, author, time)])


    def test_fixed(self):
        """
        If a ticket change indicates the ticket goes from not-fixed to fixed,
        L{getActions} emits a C{"fixed"} action.
        """
        ticket = 1234
        author = u'alice'
        time = 4321
        row = (ticket, time, author, u'resolution', None, u'fixed')
        actions = list(getActions([row]))
        self.assertEquals(actions, [(FIXED, ticket, author, time)])


    def test_reopened(self):
        """
        If a ticket change indicates the ticket goes from fixed to not-fixed,
        L{getActions} emits a C{"reopened"} action with the author set to the
        user who reopened the ticket.

        The reopen action should grant a small number of points to the reopener,
        as a reward for their diligence.
        """
        ticket = 1234
        another = u'bob'
        time = 4321
        rows = [
            (ticket, time, another, u'resolution', u'fixed', None),
        ]

        actions = list(getActions(rows))
        self.assertEquals(actions, [(REOPENED, ticket, another, time)])


    def test_badreview(self):
        """
        If a ticket change indicates the ticket goes from fixed to not-fixed and
        their is an earlier review event for that ticket, L{getActions} emits a
        C{"bad-review"} action with the user set to the reviewer who failed to
        do a good review.
        """
        ticket = 7654
        reviewer = u'alice'
        time = 827572
        reopener = u'bob'

        rows = [
            (ticket, time, reviewer, u'keywords', u'review', u''),
            (ticket, time + 10, reopener, u'resolution', u'fixed', None)]

        actions = list(getActions(rows))
        self.assertEquals(
            actions,
            [(REVIEWED, ticket, reviewer, time),
             (REOPENED, ticket, reopener, time + 10),
             (BAD_REVIEW, ticket, reviewer, time + 10)])


    def test_nonCommitterFixCommentAfter(self):
        """
        The comments field includes the "real" author and reviewer of a change;
        if a committer does a review and commits the fix, most of the credit for
        the fix should go to the author of the fix.  Therefore a 'comment'
        change with a matching 'resolution' change will result in a FIXED
        action and the committer will get credit for a separate FIX_COMMIT
        action.
        """
        ticket = 8283
        time = 4321
        reviewer = u'alice'
        submitter = u'bob'

        rows = [(ticket, time, reviewer, u'resolution', u'', u'fixed'),
                (ticket, time, reviewer, u'comment', u'200',
                 u'Author: %s\nReviewer: %s\n' % (submitter, reviewer))]

        self.assertEquals(list(getActions(rows)),
                          [(FIXED, ticket, submitter, time),
                           (FIX_COMMIT, ticket, reviewer, time)])


class HelperTests(TestCase):

    def test_parseAuthorFromComment(self):
        author = u'carol'
        self.assertEquals(parseAuthorFromComment(
            "Fix blah blah\n"
            "Author: %s\n"
            "Reviewer: dennis\n"
            "\n"
            "More blah.\n" % (author,)),
            [author])


    def test_parseAuthorsFromComment(self):
        author = u'carol'
        self.assertEquals(parseAuthorFromComment(
            "Fix blah blah\n"
            "Authors: %s\n"
            "Reviewer: dennis\n"
            "\n"
            "More blah.\n" % (author,)),
            [author])


    def test_multiAuthorFromComment(self):

        authors = [u'carol', u'bob']
        self.assertEquals(parseAuthorFromComment(
            "Fix blah blah\n"
            "Authors: %s\n"
            "Reviewer: dennis\n"
            "\n"
            "More blah.\n" % (u', '.join(authors),)),
            authors)



class GetScoresTests(TestCase):
    def test_reviewed(self):
        """
        If a user reviews a ticket, the dictionary L{getscores} returns includes
        that user's name as a key mapped to the number of points users get for
        reviewing a ticket.
        """
        expected = factors[REVIEWED].points
        aliceOneReview = [(REVIEWED, 1234, u'zlice', 4321)]
        first = getscores(aliceOneReview)
        self.assertEquals([(expected, u'zlice')], first)
        second = getscores(aliceOneReview +
                           [(REVIEWED, 1234, u'bob', 4322 + n) for n in (1, 2)])
        self.assertEquals([(expected * 2, u'bob'),
                           (expected, u'zlice')],
                          second)

